\ SOL editor
\ By Nils M Holm, 2001,2004,2010

: editor-dict ;   dict>

 true var e.indent  \ auto indent
false var e.mbell   \ margin bell

    0 var eblk  \ current block
    0 var ex    \ cursor x position
    0 var ey    \ cursor y position

false var emodi    \ block modified
false var emsg     \ message posted
    0 var e#buf    \ lines buffered
    0 var e>mark   \ position of mark

0 var ebuf b/r allot   \ single line cut/paste buffer
0 var elast b/r allot  \ overflow buffer

: vert-line ( x -- )
  16 1 do  dup i cursor  [char] | emit  loop  drop ;

: horz-line ( y -- )
  0 swap cursor  [char] + emit
  64 for  [char] - emit  next 
  [char] + emit ;

: frame ( -- )
  0 vert-line  65 vert-line  0 horz-line  17 horz-line
  65 e>mark @ 1+ cursor  [char] < emit ;

: .line ( n -- ) >r  1 r@ 1+ cursor
  dtb  r@ b/r * +  b/r type  rdrop ;

: .lines ( n -- ) 15 swap do  i .line  loop ;

: redraw ( -- ) page  frame  0 .lines ;

: eline ( -- a ) dtb ey @ b/r * + ;

: erest ( -- a ) eline  ex @ + ;

: esync ( -- ) ex @ 1+  ey @ 1+ cursor ;

: draweol ( -- ) eline ex @ +  b/r ex @ - type ;

: efwd ( -- ) ex @ 1+  b/r mod  ex ! ;

: eback ( -- ) ex @ 1-  dup 0< if  b/r + then  ex ! ;

: edown ( -- ) ey @ 1+  r/blk mod  ey ! ;

: eup ( -- ) ey @ 1-  dup 0< if  r/blk + then  ey ! ;

: elen ( a -- n ) 0  b/r for  over i + 1- c@  bl <> if
     drop i break then  next  nip ;

: e#rest ( -- n ) b/r ex @ - ;

: ehome ( -- ) 0 ex ! ;

: eeoln ( -- ) eline  elen  b/r 1- min  ex ! ;

: echange ( -- ) true emodi ! ;

: e>buf ( -- a ) pad 130 + ;

: eclrmsg ( -- ) 1 18 cursor  clreol  0 emsg !  esync ;

: navkey ( -- k ) key  emsg @ if  eclrmsg then  case
    dup key.left =  => drop [char] h ;;
    dup key.down =  => drop [char] j ;;
    dup key.up =    => drop [char] k ;;
    dup key.right = => drop [char] l ;;
  end ;

: epost ( a n -- ) 1 18 cursor  count type  true emsg ! ;

: ?yn ( a n -- f ) epost  ."  [yn] "  key
  eclrmsg  >upper [char] Y = ;

: e?modi ( -- ) emodi @ if
    s" screen not saved. lose changes?" ?yn not else
    false then  if rdrop then ;

: estatus ( -- ) s" block #" epost  eblk ? 
  emodi @ if  s"  modified" else  s"  " then
  count type ;

: eload ( n -- ) dup eblk !  block drop  0 .lines
  lhdr drop  false emodi !  estatus ;

: esave ( -- ) update  flush  false emodi ! ;

: equit ( -- ) e?modi  discard  eblk @ eload  .ok quit ;

: enext ( -- ) e?modi  eblk @ 1+ eload ;

: eprev ( -- ) e?modi  eblk @ 1- eload ;

: el> ( -- ) e?modi  l> @ ?if  eload then ;

: e<l ( -- ) e?modi  <l @ ?if  eload then ;

: -eol ( -- f ) ex @ b/r 1- < ;

: ?eol ( -- ) -eol not  if rdrop then ;

: edelch ( -- ) -eol if  erest 1+ dup 1-  e#rest 1- cmove
    then  bl  eline b/r + 1- c!  draweol  echange ;

: ebksp ( -- ) eback  bs  edelch ;

: embell ( -- ) ex @ 0=  e.mbell @  and if  7 emit then ;

: einsert ( c -- ) -eol if  erest dup 1+  e#rest 1- cmove then
  eline ex @ + c!  draweol  efwd  echange  embell ;

: esavebot ( -- ) dtb b/blk + b/r -  elast  b/r cmove ;

: erestbot ( -- ) elast  dtb b/blk + b/r -  b/r cmove
  elast b/r blank ;

: eopen ( -- ) esavebot  dtb ey @ insert
  ey @ .lines  echange ;

: elpaste ( -- ) eopen  ebuf eline b/r cmove
  ey @ .lines  echange ;

' elpaste var do-paste  \ word for pasting yanked region

: elyank ( -- ) eline  ebuf  b/r cmove
  ['] elpaste do-paste ! ;

: elcut ( -- ) elyank  dtb ey @ delete  erestbot
  ey @ .lines  echange ;

: efirst ( -- n ) eline  b/r for
    dup c@  bl <> ?break  1+  next
  eline -  b/r mod 1-  0 max ;

: ?eob ( -- ) ey @ 1+  r/blk >=  if  rdrop then ;

: eindent ( -- ) ?eob  e.indent @  if  efirst ex ! else
    0 ex ! then  edown  eopen ;

: edup ( -- ) eopen  eline  dup b/r +  swap  b/r cmove ;

: e=rtrim ( -- ) erest  e#rest  blank ;

: e+ltrim ( -- ) erest b/r +  eline b/r +  e#rest  cmove
  eline b/r +  e#rest +  ex @ blank ;

: esplit ( -- ) ?eob  ?eol  edup  e=rtrim  e+ltrim
  ey @ .lines ;

: enewline ( -- ) esplit  edown  efirst ex ! ;

: erplch ( -- ) key dup 13 = if  drop edelch enewline  else
    >visual  dup  erest c!  emit  then  echange ;

: ?2long ( -- ) eline b/r + elen  ex @ +  b/r >=
  if rdrop then ;

: ejoin ( -- ) eeoln efwd  ?eob  ?eol  ?2long
  eline b/r +  erest  b/r ex @ -  cmove
  edown  elcut  eup  esync  draweol ;

: e#query ( -- n ) src@  tib >buf !  tib >in !  tib 10 + >lim !
  query  reread  #in >r  src!  r> ;

: egoto ( -- ) e?modi  s" goto # " epost
  e#query ?if  eload then ;

: e#buf! ( -- ) e>mark @  ey @ -  abs 1+ e#buf ! ;

: ebpaste ( -- ) e#buf @  0= ?exit  e#buf @ for
    dtb ey @ insert  next  e>buf  eline  e#buf @ b/r * cmove
  ey @ .lines  echange ;

: enormblk ( -- l h ) e>mark @  ey @  2dup min  -rot max ;

: eyank ( -- ) enormblk  2dup ey !  e>mark !
  over - 1+ b/r *  swap  b/r * dtb +  e>buf  rot cmove
  e#buf!  ['] ebpaste do-paste ! ;

: eyank ( -- ) enormblk  ey !  e>mark !  frame  e#buf!
  e>mark @ b/r * dtb +  e>buf  e#buf @ b/r *  cmove
  ['] ebpaste do-paste ! ;

: ecut ( -- ) eyank  e>mark @ ey !  e#buf @ for
    dtb ey @ delete next  ey @ .lines  echange ;

: e.stk ( -- ) 1 18 cursor clrscr ." stack: " .s ;

: makeblk ( prev next -- )
  newblk  dup >r  lblock  r> eload  echange ;

: einsblk ( -- ) lhdr not ?exit  e?modi
  <l @  eblk @  makeblk ;

: eappblk ( -- ) lhdr not ?exit  e?modi
  eblk @  l> @  makeblk ;

: edelblk ( -- ) lhdr not ?exit  e?modi  case
    l> @ => l> @ ;;  <l @ => <l @ ;;  else> eblk @  end
  eblk @  dup  lfree  release   eload ;

: enewblk ( -- ) e?modi  0 0 makeblk  0 ey !  15 ex ! ;

: h" ( -- ) escape ."  compile cr ; syntax

: ehelp ( -- ) page cr
h" Cursor:        h j k l move   0 $ home/end   ^B ^F back/fwd"
h" < > prev/next  w b word/back"  h"  "
h" Cut/Paste:     x X del/backsp mm mark        dd del line   "
h" d'm del to mk  yy yank line   y'm yank to mk p P paste     "
h" dw del word    yw yank word   D del to eol   Y yank to eol "
h"  "
h" Input:         i I input      a A append     o O open line "
h" ESC end input  r replace char J join lines"  h"  "
h" Commands:      :e edit block  :w write block :q quit       "
h" ZZ save,quit   L-block:       :i insert      :a append     "
h" :n new         :d delete"
s" press enter" epost  key  redraw  drop ;

: eexit ( -- ) key [char] Z <> ?exit esave equit ;

: | ( -- ) char , ' , ;

: |^ ( -- ) char [char] @ - , ' , ;

create emenu  | a eappblk  | d edelblk  | e egoto
  | h ehelp   | i einsblk  | n enewblk  | q equit
  | w esave   0 ,  link

: eselect ( c a -- ) begin  dup @ while  over over @ = if
      nip  cell+ @ exec  exit then  cell+ cell+ repeat  2drop ;

: elkey ( -- c )  key >lower  1 18 cursor  clreol ;

: ecolon ( n -- )  1 18 cursor
  ." edit  write  quit  help  append  insert  new  delete"
  elkey emenu eselect ;

: ^ ( -- c ) bl word  1+ c@  [char] @ -  lit ; syntax

: insmode ( -- ) begin  esync key  case
    dup ^ [ = => ex @ 0= not if  eback then  drop exit ;;
    dup ^ H = => ebksp ;;
    dup ^ M = => enewline ;;
           else> dup bl [char] ~ within if 
                   dup einsert then
  end  drop  again ;

: eapp ( -- ) efwd insmode ;

: e^app ( -- ) eeoln insmode ;

: e^ins ( -- ) efirst ex !  insmode ;

: eopln ( -- ) eindent insmode ;

: e^opln ( -- ) eopen insmode ;

: -mark ( -- f ) key [char] m <> ;

: emark ( -- ) -mark ?exit  ey @ e>mark !  frame ;

: e@ ( -- c ) erest c@ ;

: eword ( -- ) efwd
    begin  e@ bl <>  ex @  and while  efwd repeat
    begin  e@ bl =   ex @  and while  efwd repeat ;

: e<word ( -- ) eback
    begin  e@ bl =   ex @  and while  eback repeat
    begin  e@ bl <>  ex @  and while  eback repeat
    e@ bl = if efwd then ;

: ewpaste ( -- ) ebuf c@ 0= ?exit
  eline elen  ebuf c@ +  b/r > if
    s" too long" epost exit then
  erest  dup  ebuf c@ +  e#rest ebuf c@ -  cmove
  ebuf 1+  erest  ebuf c@  cmove  ey @ .lines  echange ;

: (ewyank) ( -- ) 2dup < if
    over ex !  - neg  dup >r  ebuf c!
    erest  ebuf 1+  r> cmove  else  0 ebuf c! then
  2drop  ['] ewpaste do-paste ! ;

: ewyank ( -- ) ex @  eword  ex @  (ewyank) ;

: (ewcut) ( -- ) ebuf c@ 0= ?exit
  erest  dup ebuf c@ +  swap  e#rest  ebuf c@ -  cmove
  eline b/r +  ebuf c@ -  ebuf c@ blank  ey @ .lines  echange ;

: ewcut ( -- ) ewyank  (ewcut) ;

: eeyank ( -- ) ex @  eeoln  ex @  (ewyank) ;

: eecut ( -- ) eeyank  (ewcut) ;

: epaste ( -- ) do-paste @ exec ;

: e.paste ( -- ) do-paste @ ['] ewpaste = if  efwd else
    edown then  epaste ;

: ecr ( -- ) edown  ehome ;

' [char] alias &  syntax

: dcmd ( -- ) navkey  case
    dup & d = => drop elcut ;;
    dup & w = => drop ewcut ;;
    dup & ' = => drop -mark ?exit  ecut ;;
    else> drop ;;
  end ;

: ycmd ( -- ) navkey  case
    dup & y = => drop elyank ;;
    dup & w = => drop ewyank ;;
    dup & ' = => drop -mark ?exit  eyank ;;
    else> drop ;;
  end ;

create emain   | h eback    | j edown    | k eup
  | l efwd     |^ F enext   |^ B eprev   | < e<l
  | > el>      | w eword    | b e<word   | 0 ehome
  | $ eeoln    | : ecolon
  | a eapp     | A e^app    | d dcmd     | D eecut
  | i insmode  | I e^ins    | J ejoin    | m emark
  | o eopln    | O e^opln   | p e.paste  | P epaste
  | r erplch   | x edelch   | X ebksp    | y ycmd
  | Y eeyank   | Z eexit    |^ G estatus |^ K e.stk
  |^ L redraw  |^ M ecr     0 ,  link

: cmd ( -- ) begin  esync  navkey emain eselect  again ;

: einit ( -- ) 0 ex !  0 ey !  false emodi !
  ebuf b/r blank  elast b/r blank  false emsg !
  0 e>mark !  0 e#buf ! ;

: e ( n -- ) e>buf b/blk +  top u>= if
    abort"  not enough memory" then
  einit  page  frame  eload  s" type :h for help" epost
  cmd ;

<dict editor  editor hide

