module solcmp(t3x, char, string, iostream, util);

object t[t3x], chr[char], str[string], u[util];

const	BUFLEN	= 1024;

const	KSPACE	= 0x400;
const	KSIZE	= 65536;
const	STKLEN	= 64;
const	TEXTLEN	= 64;

object	In[iostream], Out[iostream];
var	Inbuf::BUFLEN, Outbuf::BUFLEN;
var	Kmem::KSIZE;
var	Line;
var	Text::TEXTLEN+1;
var	Last;
var	Here;
var	Eof;
var	Stack::STKLEN, Sp;
var	Rejected;

error(m, n) do var b::4;
	u.writef(T3X.SYSERR, "solcmp: %d: ", [(Line)]);
	u.writef(T3X.SYSERR, m, [(n)]);
	u.writef(T3X.SYSERR, "%s", [(t.newline(b))]);
	halt 1;
end

rdch() do var c;
	if (Rejected) do
		c := Rejected;
		Rejected := 0;
		return c;
	end
	return In.rdch();
end

parse(d) do var c, i;
	d := d-> d: ' ';
	c := rdch();
	while (c = d \/ c = '\n' \/ c = '\r') do
		if (c = '\n') Line := Line + 1;
		c := rdch();
	end
	if (In.eof()) do
		Eof := 1;
		return 0;
	end
	i := 0;
	while (c \= d /\ c \= '\n' /\ c \= '\r' /\ \In.eof()) do
		if (i > TEXTLEN) do
			Text::TEXTLEN := 0;
			error("token too long: %s", Text);
		end
		Text::i := c;
		i := i + 1;
		c := rdch();
	end
	if (c = '\n') Rejected := c;
	Text::i := 0;
	return Text;
end

scomment(dlm) do var c;
	c := rdch();
	while (c \= dlm /\ \In.eof()) do
		if (In.eof()) error("EOF in comment", 0);
		if (c = '\n') Line := Line + 1;
		c := rdch();
	end
	if (c = '\n') Line := Line + 1;
	parse(0);
end

getw(x)
	return (Kmem::(x-KSPACE) << 24)
		| (Kmem::(x-KSPACE+1) << 16)
		| (Kmem::(x-KSPACE+2) << 8)
		| Kmem::(x-KSPACE+3);

putw(x, v) do
	Kmem::(x-KSPACE) := v >> 24;
	Kmem::(x-KSPACE+1) := v >> 16;
	Kmem::(x-KSPACE+2) := v >> 8;
	Kmem::(x-KSPACE+3) := v;
end

emit(x) do
	Kmem::(Here-KSPACE) := x;
	Here := Here + 1;
end

emitw(x) do
	emit(x>>24);
	emit(x>>16);
	emit(x>>8);
	emit(x);
end

emits(s) do var i;
	i := 0;
	while (s::i) do
		emit(s::i);
		i := i+1;
	end
end

screate() do var k;
	emitw(Last);
	Last := Here-4;
	parse(0);
	k := str.length(Text);
	emit(k);
	emits(Text);
	emit(k);
	parse(0);
end

byteval(s) do var n;
	ie ('0' <= s::0 /\ s::0 <= '9') n := s::0-'0';
	else ie ('A' <= s::0 /\ s::0 <= 'F') n := s::0-'7';
	else ie ('a' <= s::0 /\ s::0 <= 'f') n := s::0-'W';
	else return -1;
	if (\s::1) return n;
	n := n << 4;
	ie ('0' <= s::1 /\ s::1 <= '9') n := n | s::1-'0';
	else ie ('A' <= s::1 /\ s::1 <= 'F') n := n | s::1-'7';
	else ie ('a' <= s::1 /\ s::1 <= 'f') n := n | s::1-'W';
	else return -1;
	if (s::2) return -1;
	return n;
end

primitive() do var b;
	screate();
	while (str.comp(Text, "next;")) do
		b := byteval(Text);
		ie (b >= 0) do
			emit(b);
			parse(0);
		end
		else ie (\str.comp(Text, "next,")) do
			emit(0x82);	! NEXT
			parse(0);
		end
		else ie (\str.comp(Text, "\\")) do
			scomment('\n');
		end
		else do
			error("syntax error in :H at %s", Text);
		end
	end
	emit(0x82);	! NEXT
	parse(0);
end

numeric(s, v) do var n, i;
	n := 0;
	i := 0;
	while (s::i) do
		if (\chr.digit(s::i)) return 0;
		n := n*10 + s::i - '0';
		i := i+1;
	end
	v[0] := n;
	return 1;
end

push(n) do
	if (Sp >= STKLEN) error("stack overflow", 0);
	Stack[Sp] := n;
	Sp := Sp+1;
end

pop() do
	if (Sp < 1) error("stack underflow", 0);
	Sp := Sp-1;
	return Stack[Sp];
end

swap() do var x;
	if (Sp < 2) error("stack underflow", 0);
	x := Stack[Sp-1];
	Stack[Sp-1] := Stack[Sp-2];
	Stack[Sp-2] := x;
end

_lookup(s, adj) do var p, k;
	k := str.length(s);
	p := Last;
	while (p) do
		if (	(Kmem::(p-KSPACE+4) & 127) = k /\
			\t.memcomp(@Kmem::(p-KSPACE+5), s, k)
		) do
			if (adj) p := p + (Kmem::(p-KSPACE+4) & 127) + 6;
			return p;
		end
		p := getw(p);
	end
	return 0;
end

lookup(s, adj) do var p;
	p := _lookup(s, adj);
	if (p = 0) error("undefined: %s", s);
	return p;
end

sfind() do var a;
	parse(0);
	a := lookup(Text, 1);
	parse(0);
	push(a);
end

scall(a) do
	emitw(a);	! CALL implied
end

svar() do
	screate();
	scall(lookup("(var)", 1));
	emitw(pop());
end

sconst() do
	screate();
	scall(lookup("(const)", 1));
	emitw(pop());
end

sarray() do var i, k;
	screate();
	scall(lookup("(array)", 1));
	k := pop();
	for (i=0, k) emitw(0);
end

sallot() do var i, n;
	n := pop();
	i := 0;
	while (i<n) do
		emit(0);
		i := i+1;
	end
	parse(0);
end

scomma() do
	emitw(pop());
	parse(0);
end

sccomma() do
	emit(pop());
	parse(0);
end


sbang() do var a, n;
	a := pop();
	n := pop();
	putw(a, n);
	parse(0);
end

sctick() do
	parse(0);
	emitw(lookup("(lit)", 1));
	emitw(lookup(Text, 1));
	parse(0);
end

scompile() do
	parse(0);
	emitw(lookup("(lit)", 1));
	emitw(lookup(Text, 1));
	parse(0);
	emitw(lookup(",", 1));
end

sclit(n) do
	emitw(lookup("(lit)", 1));
	emitw(n);
	parse(0);
end

sif() do
	emitw(lookup("(bf)", 1));
	push(Here);
	emitw(0);
	parse(0);
end

sthen() do var a;
	a := pop();
	putw(a, Here);
	parse(0);
end

selse() do
	emitw(lookup("(br)", 1));
	push(Here);
	emitw(0);
	swap();
	sthen();
end

srepeat() do var a;
	emitw(lookup("(br)", 1));
	a := pop();
	emitw(pop());
	putw(a, Here);
	parse(0);
end

suntil() do
	emitw(lookup("(bf)", 1));
	emitw(pop());
	parse(0);
end

sagain() do
	emitw(lookup("(br)", 1));
	emitw(pop());
	parse(0);
end

sfor() do
	emitw(lookup(">r", 1));
	push(Here);
	parse(0);
end

snext() do
	emitw(lookup("(next)", 1));
	emitw(lookup("(bf)", 1));
	emitw(pop());
	parse(0);
end

sdo() do
	emitw(lookup("(do)", 1));
	push(Here);
	parse(0);
end

sloop() do
	emitw(lookup("(loop)", 1));
	emitw(lookup("(bf)", 1));
	emitw(pop());
	parse(0);
end

scase() do
	push(0);
	parse(0);
end

sarrow() do
	emitw(lookup("(bf)", 1));
	push(Here);
	emitw(0);
	parse(0);
end

s2semi() do var a;
	a := pop();
	emitw(lookup("(br)", 1));
	push(Here);
	emitw(0);
	putw(a, Here);
	parse(0);
end

s_end() do var a;
	while (1) do
		a := pop();
		if (a = 0) leave;
		putw(a, Here);
	end
	parse(0);
end

scchar() do
	emitw(lookup("(lit)", 1));
	parse(0);
	emitw(Text::0);
	parse(0);
end

schex() do var x;
	emitw(lookup("(lit)", 1));
	parse(0);
	str.parse(Text, "%X", [(@x)]);
	emitw(x);
	parse(0);
end

hex() do var x;
	parse(0);
	str.parse(Text, "%X", [(@x)]);
	push(x);
	parse(0);
end

sabort() do
	parse('"');
	emitw(lookup("(abort\")", 1));
	emit(str.length(Text));
	emits(Text);
	parse(0);
end

sdotquote() do
	parse('"');
	emitw(lookup("(.\")", 1));
	emit(str.length(Text));
	emits(Text);
	parse(0);
end

sstrquote() do
	parse('"');
	emitw(lookup("(s\")", 1));
	emit(str.length(Text));
	emits(Text);
	parse(0);
end

scolon() do var a, n, esc;
	esc := 0;
	screate();
	scall(lookup("(:)", 1));
	while (esc \/ str.comp(Text, ";")) do
		if (Eof) error("EOF in : definition", 0);
		ie (esc) ;
		else ie (\str.comp(Text, "\\")) scomment('\n');
		else ie (\str.comp(Text, "(")) scomment(')');
		else ie (\str.comp(Text, "escape")) do esc := 1; parse(0); end
		else ie (\str.comp(Text, "[']")) sctick();
		else ie (\str.comp(Text, "compile")) scompile();
		else ie (\str.comp(Text, "if")) sif();
		else ie (\str.comp(Text, "else")) selse();
		else ie (\str.comp(Text, "then")) sthen();
		else ie (\str.comp(Text, "begin")) do push(Here); parse(0); end
		else ie (\str.comp(Text, "while")) sif();
		else ie (\str.comp(Text, "repeat")) srepeat();
		else ie (\str.comp(Text, "until")) suntil();
		else ie (\str.comp(Text, "again")) sagain();
		else ie (\str.comp(Text, "for")) sfor();
		else ie (\str.comp(Text, "next")) snext();
		else ie (\str.comp(Text, "do")) sdo();
		else ie (\str.comp(Text, "loop")) sloop();
		else ie (\str.comp(Text, "case")) scase();
		else ie (\str.comp(Text, "=>")) sarrow();
		else ie (\str.comp(Text, "else>")) parse(0);
		else ie (\str.comp(Text, ";;")) s2semi();
		else ie (\str.comp(Text, "end")) s_end();
		else ie (\str.comp(Text, "[char]")) scchar();
		else ie (\str.comp(Text, "[hex]")) schex();
		else ie (\str.comp(Text, "abort\"")) sabort();
		else ie (\str.comp(Text, ".\"")) sdotquote();
		else ie (\str.comp(Text, "s\"")) sstrquote();
		else esc := 2;
		if (esc) do
			a := _lookup(Text, 0);
			ie (a = 0) do
				ie (numeric(Text, @n)) do
					sclit(n);
				end
				else do
					lookup(Text, 0);
				end
			end
			else do
				if (esc \= 1 /\ Kmem::(a-KSPACE+4) & 128)
					error("unknown syntax word: %s", Text);
				a := a + (Kmem::(a-KSPACE+4) & 127) + 6;
				emitw(a);
				parse(0);
			end
			esc := 0;
		end
	end
	emitw(lookup("exit", 1));
	parse(0);
end

salias() do
	screate();
	emit(0x8a);	! REDIR
	emitw(pop());
end

ssyntax() do
	Kmem::(Last-KSPACE+4) := Kmem::(Last-KSPACE+4) | 128;
	parse(0);
end

sdefer() do
	screate();
	scall(lookup("(:)", 1));
	emitw(lookup("exit", 1));	! xxx should be (DEFER)
	emitw(lookup("exit", 1));
end

sis() do var a, d;
	sfind();
	d := pop()+4;
	a := pop();
	putw(d, a);
end

stolink() do var a;
	a := pop() - 1;
	a := a - Kmem::(a-KSPACE) - 5;
	push(a);
	parse(0);
end

neg() do
	Stack[Sp-1] := -Stack[Sp-1];
	parse(0);
end

minus() do
	Stack[Sp-2] := Stack[Sp-2] - Stack[Sp-1];
	Sp := Sp-1;
	parse(0);
end

pushvar(s) do
	push(lookup(s, 1) + 4);
	parse(0);
end

findvar(s) do var a, va, ca, d;
	a := lookup(s, 1);
	d := getw(a);
	va := lookup("(var)", 1);
	ie (d = va) do
		push(a+4);
	end
	else do
		ca := lookup("(const)", 1);
		ie (d = ca) do
			d := getw(a+4);
			push(d);
		end
		else do
			error("syntax error at %s", s);
		end
	end
	parse(0);
end

compile() do var n;
	parse(0);
	while (\Eof) do
		ie (\str.comp(Text, "\\d")) scomment('\n');
		else ie (\str.comp(Text, "\\")) scomment('\n');
		else ie (\str.comp(Text, "(")) scomment(')');
		else ie (\str.comp(Text, ":h")) primitive();
		else ie (numeric(Text, @n)) do push(n); parse(0); end
		else ie (\str.comp(Text, "var")) svar();
		else ie (\str.comp(Text, "const")) sconst();
		else ie (\str.comp(Text, "array")) sarray();
		else ie (\str.comp(Text, "value")) sconst();
		else ie (\str.comp(Text, "allot")) sallot();
		else ie (\str.comp(Text, "'")) sfind();
		else ie (\str.comp(Text, ",")) scomma();
		else ie (\str.comp(Text, "c,")) sccomma();
		else ie (\str.comp(Text, "!")) sbang();
		else ie (\str.comp(Text, ":")) scolon();
		else ie (\str.comp(Text, "alias")) salias();
		else ie (\str.comp(Text, "syntax")) ssyntax();
		else ie (\str.comp(Text, "defer")) sdefer();
		else ie (\str.comp(Text, "is")) sis();
		else ie (\str.comp(Text, "hex")) hex();
		else ie (\str.comp(Text, "here")) do push(Here); parse(0); end
		else ie (\str.comp(Text, "h")) pushvar(Text);
		else ie (\str.comp(Text, ">link")) stolink();
		else ie (\str.comp(Text, "neg")) neg();
		else ie (\str.comp(Text, "-")) minus();
		else findvar(Text);
	end
	Out.write(Kmem, Here-KSPACE);
end

do var b::4;
	In.create(T3X.SYSIN, Inbuf, BUFLEN, IOSTREAM.FREAD);
	Out.create(T3X.SYSOUT, Outbuf, BUFLEN, IOSTREAM.FWRITE);
	Line := 1;
	Last := 0;
	Here := KSPACE;
	Eof := 0;
	Sp := 0;
	Rejected := 0;
	compile();
	u.writef(T3X.SYSERR, "size = %ud%s", [(Here-KSPACE, t.newline(b))]);
	Out.flush();
end
