module solfs(t3x, char, string, memory, iostream, util);

object	t[t3x], chr[char], str[string], mem[memory], u[util];

struct FILE = FNAME, FONAME, FADDR, FBLOCK0, FBLOCKS;

var	Files[16];

const	BUFLEN		= 1024;
const	POOLSIZE	= 4096;
const	STRLEN		= 256;
const	USERBLOCKS	= 132;
const	DEVSIZE		= 8192;

object	Out[iostream];
var	Outbuf::BUFLEN;
var	Pool::POOLSIZE;
var	Nfiles;
var	Free;
var	Freelist::1024;
var	Block::1024;
var	InitBlock::STRLEN;

const	BLKSIZE	= 1024;
const	KSIZE	= 64*BLKSIZE;

error(m, n) do var b::4;
	u.writef(T3X.SYSERR, "SOLFS: ", 0);
	u.writef(T3X.SYSERR, m, [(n)]);
	u.writef(T3X.SYSERR, "%s", [(t.newline(b))]);
	halt 1;
end

append(name) do
	var	fd;
	var	buf::1024;
	var	k;
	var	n;

	n := 0;
	fd := t.open(name, T3X.OREAD);
	if (fd < 0) error("no such file: %s", name);
	while (1) do
		k := t.read(fd, buf, 1024);
		if (k < 1) leave;
		n := n + k;
		Out.write(buf, k);
	end
	t.close(fd);
	return n;
end

writes(s) do var i, k;
	k := str.length(s);
	Out.write(s, k);
	for (i=k, 64) Out.wrch('\s');
end

appendText(name) do
	object	src[iostream];
	var	fbuf::256, buf::67;
	var	k;
	var	n;

	n := 0;
	if (src.open(name, fbuf, 256, IOSTREAM.FREAD) = %1)
		error("no such file: %s", name);
	while (1) do
		k := src.reads(buf, 66);
		if (src.eof()) leave;
		if (k /\ buf::(k-1) = '\n') do
			buf::(k-1) := 0;
			if (k>1 /\ buf::(k-2) = '\r') do
				buf::(k-2) := 0;
			end
		end
		writes(buf);
		n := n + 64;
	end
	src.close();
	return n;
end

fill(n, lim, ch) do
	var	k;

	if (n > lim) error("limit check (lim=%d)", lim);
	while (n<lim) do
		Out.wrch(ch);
		n := n+1;
	end
	return lim;
end

alloc(n) do var p;
	p := mem.alloc(n);
	if (p = 0) error("out of memory", 0);
	return p;
end

strsave(s) do var p;
	p := alloc(str.length(s)+1);
	str.copy(p, s);
	return p;
end

locase(s) do
	var	new;
	var	i, k;

	k := str.length(s)+1;
	new := alloc(k);
	for (i=0, k)
		ie (chr.lower(s::i))
			new::i := chr.lcase(s::i);
		else
			new::i := s::i;
	return new;
end

collect(name, b0) do
	object	in[iostream];
	var	inbuf::BUFLEN;
	var	buf::256, b::4;
	var	nr, nb;
	var	k;
	var	f;

	if (Nfiles >= 15) error("too many files (%s)", name);
	if (in.open(name, inbuf, BUFLEN, IOSTREAM.FREAD) = %1)
		error("no such file: %s", name);
	in.reads(buf, 80);
	nr := 0;
	while (\in.eof()) do
		if (\(nr mod 16)) nr := nr+1;
		k := str.length(buf);
		if (k > 65 \/ buf::(k-1) \= '\n') do
			str.format(buf, "%s: line too long: %%d", [(name)]);
			error(buf, nr+1);
		end
		nr := nr+1;
		in.reads(buf, 80);
	end
	in.close();
	nb := (nr+15) / 16;
	Files[Nfiles] := alloc(FILE * t.bpw());
	f := Files[Nfiles];
	f[FNAME] := locase(name);
	f[FONAME] := strsave(name);
	f[FBLOCKS] := nb;
	f[FADDR] := Free;
	u.writef(T3X.SYSERR, "%3:\sd %s%s", [(nb, f[FNAME], t.newline(b))]);
	Nfiles := Nfiles + 1;
	Free := Free + nb;
	return nb;
end

copyin(fdesc) do
	object	in[iostream];
	var	inbuf::BUFLEN;
	var	ibuf::66, obuf::65, buf::65;
	var	nr, nb, k, i;
	var	prev, this, next;

	if (in.open(fdesc[FONAME], inbuf, BUFLEN, IOSTREAM.FREAD) = %1)
		error("%s: OOPS - file vanished?", fdesc[FONAME]);
	nr := 0;
	nb := 0;
	prev := 0;
	this := fdesc[FADDR];
	next := this+1;
	in.reads(ibuf, 65);
	while (\in.eof()) do
		if (\(nr mod 16)) do
			nb := nb+1;
			if (nb >= fdesc[FBLOCKS]) next := 0;
			str.format(buf, "(l %5:0d %5:0d %s )",
				[(prev, next, fdesc[FNAME])]);
			str.format(obuf, "%64s", [(buf)]);
			Out.write(obuf, 64);
			prev := this;
			this := next;
			next := next+1;
			nr := nr+1;
		end
		k := str.length(ibuf);
		ibuf::(k-1) := 0;
		str.format(obuf, "%64s", [(ibuf)]);
		Out.write(obuf, 64);
		in.reads(ibuf, 65);
		nr := nr+1;
	end
	str.format(obuf, "%64s", [""]);
	while (nr mod 16) do
		Out.write(obuf, 64);
		nr := nr+1;
	end
	in.close();
	return nb;
end

bitset(a, n) do
	var	mask, off;
	var	i;

	off := n ./ 8;
	i := n mod 8;
	mask := 1;
	while (i) do
		mask := mask << 1;
		i := i-1;
	end
	a::off := a::off | mask;
end

mkfs() do
	var	z, zz;
	var	i;
	var	used;
	var	dbuf::65;
	var	fp;
	var	path::256;
	var	b::3;

	z := 0;
	! collect files 
	if (t.getarg(1, InitBlock, 256) < 1)
		error("no init block", 0);
	used := 0;
	i := 2;
	while (t.getarg(i, path, 256) >= 0) do
		used := used + collect(path, used+USERBLOCKS);
		i := i+1;
	end

	! Write boot block 
	t.memfill(Block, 0, 1024);
	Block::3 := 1;
	Block::7 := KSIZE/BLKSIZE;
	Out.write(Block, 1024);
	z := BLKSIZE;

	! Write kernel images
	zz := z + append("solimage");
	z := fill(zz, z+KSIZE, 0);
	zz := z + append("solimage");
	z := fill(zz, z+KSIZE, 0);

	! write free list 
	t.memfill(Freelist, 0, 1024);
	for (i=0, used+USERBLOCKS) bitset(Freelist, i);
	for (i=DEVSIZE, 8192) bitset(Freelist, i);
	Out.write(Freelist, 1024);
	z := z + BLKSIZE;

	! Write init block
	writes("( initialization block )");
	zz := z + 64 + appendText(InitBlock);
	z := fill(zz, z+BLKSIZE, '\s');

	! write base directory
	writes("(l 00000 00000 table of contents )");
	zz := z + 64;
	for (i=0, Nfiles) do
		fp := Files[i];
		str.format(dbuf, "%5d const %30s%22s", [(
			fp[FADDR],
			fp[FNAME],
			"")]);
		writes(dbuf);
		zz := zz + 64;
	end
	z := fill(zz, z+BLKSIZE, '\s');

	! copy files to image 
	for (i=0, Nfiles)
		z := z + copyin(Files[i]) * BLKSIZE;

	! fill device
	zz := z;
	z := fill(z, BLKSIZE*DEVSIZE, 0);
	Out.flush();

	u.writef(T3X.SYSERR, "blocks: %d total, %d used, %d free%s",
		[(DEVSIZE, zz/BLKSIZE, DEVSIZE-zz/BLKSIZE, t.newline(b))]);
end

do
	Out.create(T3X.SYSOUT, Outbuf, BUFLEN, IOSTREAM.FWRITE);
	mem.init(Pool, POOLSIZE);
	Nfiles := 0;
	Free := USERBLOCKS;
	mkfs();
end
